'''
SYNOPSIS

    class MongoDocSpec

DESCRIPTION

    Specify documents of MongoDB.

    indexes is list of pymongo.IndexModel.

REFERENCES

    pymongo.IndexModel
        https://api.mongodb.com/python/current/api/pymongo/operations.html#\
        pymongo.operations.IndexModel
    PyMongo - Create document index
        https://api.mongodb.com/python/current/api/pymongo/collection.html#\
        pymongo.collection.Collection.create_indexes
'''

from clink.com import com, Component


class MongoDocSpec():
    def __init__(self, name, indexes):
        self._name = name
        self._indexes = indexes

    @property
    def name(self):
        return self._name

    @property
    def indexes(self):
        return self._indexes


@com()
class MongoConf(Component):
    def __init__(self, dburl, dbname):
        self._dburl = dburl
        self._dbname = dbname

    @property
    def dburl(self):
        return self._dburl

    @property
    def dbname(self):
        return self._dbname



from clink.com.injector import CLINK_COM_ATTR
from clink.mime.type import MIME_JSON


def path(path):
    def new_fn(target_fn):
        target_fn._path = path
        if CLINK_COM_ATTR in dir(target_fn):
            target_fn.__clink['route_path'] = path
        else:
            target_fn.__clink = {'route_path': path}
        return target_fn
    return new_fn


def map(method, path, req_type):
    def decorator_fn(target_fn):
        raw_route = (method, path, req_type)
        if CLINK_COM_ATTR in dir(target_fn):
            target_fn.__clink['raw_route'] = raw_route
        else:
            target_fn.__clink = {'raw_route': raw_route}
        return target_fn
    return decorator_fn


def get(path):
    return map('get', path, None)


def post(path, req_type=MIME_JSON):
    return map('post', path, req_type)


def put(path, req_type=MIME_JSON):
    return map('put', path, req_type)


def patch(path, req_type=MIME_JSON):
    return map('patch', path, req_type)


def delete(path):
    return map('delete', path, None)



from pymongo import ASCENDING, IndexModel

from clink.service.mongo import MongoDocSpec

from clink.com import com
from clink.type.com import Service
from clink.service.mongo import MongoService

_ROOT_NAME = 'root'
_ACC_DOCNAME = 'account'
_GRP_DOCNAME = 'group'
_RPWD_DOCNAME = 'rpwd'
_ACCTMP_DOCNAME = 'acctmp'

_ACC_IND_1 = IndexModel([('name', ASCENDING)], unique=True)
_ACC_IND_2 = IndexModel([('email', ASCENDING)], unique=True)
_ACC_DOCSPEC = MongoDocSpec(_ACC_DOCNAME, [_ACC_IND_1, _ACC_IND_2])

_GRP_IND_1 = IndexModel([('name', ASCENDING)], unique=True)
_GRP_DOCSPEC = MongoDocSpec(_GRP_DOCNAME, [_GRP_IND_1])

_RPWD_DOCSPEC = MongoDocSpec(_RPWD_DOCNAME, [])

_ACCTMP_DOCSPEC = MongoDocSpec(_ACCTMP_DOCNAME, _ACC_DOCSPEC.indexes)

_DOC_SPECS = [_ACC_DOCSPEC, _GRP_DOCSPEC, _RPWD_DOCSPEC, _ACCTMP_DOCSPEC]


@com(MongoService)
class AuthDbService(Service):
    def __init__(self, mongo_sv):
        mongo_sv.use_docspecs(_DOC_SPECS)



from pymongo import ASCENDING, IndexModel

from ..db import MongoDocSpec

from .iauth import IAuth
from .accmgr import AccMgr
from .oauth import OAuth

_ROOT_NAME = 'root'
_ACC_DOCNAME = 'account'
_GRP_DOCNAME = 'group'
_RPWD_DOCNAME = 'rpwd'
_ACCTMP_DOCNAME = 'acctmp'

_ACC_IND_1 = IndexModel([('name', ASCENDING)], unique=True)
_ACC_IND_2 = IndexModel([('email', ASCENDING)], unique=True)
_ACC_DOCSPEC = MongoDocSpec(_ACC_DOCNAME, [_ACC_IND_1, _ACC_IND_2])

_GRP_IND_1 = IndexModel([('name', ASCENDING)], unique=True)
_GRP_DOCSPEC = MongoDocSpec(_GRP_DOCNAME, [_GRP_IND_1])

_RPWD_DOCSPEC = MongoDocSpec(_RPWD_DOCNAME, [])

_ACCTMP_DOCSPEC = MongoDocSpec(_ACCTMP_DOCNAME, _ACC_DOCSPEC.indexes)

_DOC_SPECS = [_ACC_DOCSPEC, _GRP_DOCSPEC, _RPWD_DOCSPEC, _ACCTMP_DOCSPEC]

class Auth(IAuth):
    def __init__(
        self, dbnode, root_pwd, root_email, jwt_key,
        token_time=4*3600, rtoken_time=30*24*3600
    ):
        dbnode.use_docspecs(_DOC_SPECS)

        acc_doc, grp_doc = dbnode.docs(_ACC_DOCNAME, _GRP_DOCNAME)
        rpwd_doc, acctmp_doc = dbnode.docs(_RPWD_DOCNAME, _ACCTMP_DOCNAME)
        self._accmgr = AccMgr(acc_doc, grp_doc, rpwd_doc, acctmp_doc)
        self._auth = OAuth(acc_doc, grp_doc, jwt_key, token_time, rtoken_time)

        root_acc = self._accmgr.find_name(_ROOT_NAME)
        if root_acc is None:
            self._accmgr.create(_ROOT_NAME, root_pwd, root_email)

    @property
    def accmgr(self):
        return self._accmgr

    @property
    def auth(self):
        return self._auth



from clink.iface import IPipeHandler
from clink.com.marker import com
from clink.type.com import AppResHandler


@com()
class ResCorsHandler(AppResHandler, IPipeHandler):
    def handle(self, req, res):
        if req.method.lower() != 'option':
            return
        res.header['Access-Control-Allow-Origin'] = '*'
        res.header['Access-Control-Allow-Methods'] = \
            'GET,POST,PUT,DELETE,OPTIONS'
        res.header['Access-Control-Allow-Headers'] = \
            'Authorization,Content-Type'



'''
SYNOPSIS

    class Route

DESCRIPTION

    Decorate function to add it's function to routing.
'''

import os

from ..mime.type import MIME_JSON
from ..etc import URL_SLASH
from .error import RoutePathError
from .type import RouteSpec


class Route():
    def __init__(self, base_path):
        if not self.verify_path(base_path):
            raise RoutePathError(base_path)
        self.base_path = base_path
        self.specs = []

    def add_spec(self, method, path, req_type, handler):
        if not self.verify_path(path):
            raise RoutePathError(path)

        abs_path = os.path.join(self.base_path, path) 
        spec = RouteSpec(method, abs_path, handler, req_type)
        self.specs.append(spec)


    def verify_path(self, path):
        if len(path) > 0:
            if path[0] == URL_SLASH:
                return False
        return True



class ComInvalidError(Exception):
    def __init__(self, com_type):
        self._msg = com_type.__name__

    def __str__(self):
        return self._msg


class CircleComError(Exception):
    def __init__(self, com_types):
        self._msg = ', '.join([t.__name__ for t in com_types])

    def __str__(self):
        return self._msg


class ComExistError(Exception):
    def __init__(self, com_type):
        self._msg = com_type.__name__

    def __str__(self):
        return self._msg


class ComCreationError(Exception):
    def __init__(self, com_type, args):
        self._msg = 'type=%s, args=%s' % (com_type, args)


    def __str__(self):
        return self._msg



import random
from hashlib import sha224
from string import ascii_lowercase, ascii_uppercase, digits

_PWD_CHARS = ascii_lowercase + ascii_uppercase + digits
_CODE_CHARS = ascii_uppercase


def hash_pwd(password):
    return sha224(password.encode('utf-8')).hexdigest()


def rand_pwd():
    return ''.join(random.sample(_PWD_CHARS, 6))


def rand_code():
    a = ''.join(random.sample(_CODE_CHARS, 4))
    b = ''.join(random.sample(_CODE_CHARS, 4))
    c = ''.join(random.sample(_CODE_CHARS, 4))
    d = ''.join(random.sample(_CODE_CHARS, 4))

    return '-'.join([a, b, c, d])



import clink

from clink import com
from clink.type.com import Controller
from clink.ske.auth import ctl


class App(clink.App):
    def __init__(self, app_conf, mongo_conf, auth_conf):
        super().__init__(app_conf)

        self.injector.add_inst(mongo_conf)
        self.injector.add_inst(auth_conf)

        ctl_coms = com.find(ctl, Controller)
        self.injector.add_coms(ctl_coms)



'''
SYNOPSIS

    class AccMgr(acc_doc, grp_doc)
        create(self, name, password, email, phone=None)
        find_id(self, id)
        find_name(self, name)
        remove(self, id)
        change_pwd(self, id, new_pwd)
        reset_pwd(self, email)
        add_group(self, group_name)
        del_group(self, group_name)
        add_to_group(self, acc_id, group_name)
        del_fm_group(self, acc_id, group_name)

DESCRIPTION

    Manage accounts and groups.
'''

from datetime import datetime, timedelta
from time import time
from string import ascii_lowercase, ascii_uppercase, digits

from clink.type.com import Service
from clink.com import com
from clink.service.mongo import MongoService

from .error import AccountNotExist, GroupExist, GroupNotExist, \
                   CodeNotExistError, CodeExpiredError, AccountExistError, \
                   EmailExistError
from .authdb_sv import AuthDbService
from .type import AuthConf
from .util import hash_pwd, rand_pwd, rand_code

_ACT_REGISTERED = 'REGISTERED'
_ACT_CHANGE_PWD = 'CHANGE_PWD'
_ACT_RESET_PWD = 'RESET_PWD'
_ACT_ADD_TO_GRP = 'ADD_TO_GRP'
_ACT_RM_FRM_GRP = 'RM_FRM_GRP'


@com(MongoService, AuthDbService, AuthConf)
class AccService(Service):
    _pwd_chars = ascii_lowercase + ascii_uppercase + digits

    def __init__(self, mongo_sv, authdb_sv, auth_conf):
        pass
        self._acc_doc = mongo_sv.doc('account')
        self._grp_doc = mongo_sv.doc('group')
        self._rpwd_doc = mongo_sv.doc('rpwd')
        self._acctmp_doc = mongo_sv.doc('acctmp')

        self.rpwd_time = 3600
        self.create_time = 3600

        root_acc = self.find_name('root')
        if root_acc is None:
            self.create('root', auth_conf.root_pwd, auth_conf.root_email)

    def create(self, name, password, email, phone=None):
        account = {
            'name': name,
            'hashpwd': hash_pwd(password),
            'email': email,
            'phone': phone,
            'groups': [],
            'created_date': datetime.utcnow(),
            'modified_date': datetime.utcnow(),
            'last_action': _ACT_REGISTERED
        }
        result = self._acc_doc.insert_one(account)
        return result.inserted_id

    def mk_creation(self, name, password, email, phone=None):
        if self._acc_doc.find_one({'name': name}) is not None:
            raise AccountExistError(name)
        if self._acc_doc.find_one({'email': email}) is not None:
            raise EmailExistError(email)
        if phone is not None:
            if self._acc_doc.find_one({'phone': phone}) is not None:
                raise PhoneExistError(phone)

        datetime_now = datetime.utcnow().timestamp()
        self._acctmp_doc.delete_many({'_expired_date': {'$lt': datetime_now}})

        creation_code = rand_code()
        expired_date = datetime.utcnow() + timedelta(hours=self.create_time)
        acctmp = {
            'name': name,
            'hashpwd': hash_pwd(password),
            'email': email,
            'phone': phone,
            'groups': [],
            'created_date': datetime.utcnow(),
            'modified_date': datetime.utcnow(),
            'last_action': _ACT_REGISTERED,

            '_expired_date': expired_date.timestamp(),
            '_creation_code': creation_code
        }
        self._acctmp_doc.insert_one(acctmp)

        return creation_code

    def confirm_creation(self, code):
        acctmp = self._acctmp_doc.find_one({'_creation_code': code})
        if acctmp is None:
            raise CodeNotExistError(code)
        if acctmp['_expired_date'] < datetime.utcnow().timestamp():
            raise CodeExpiredError(code)
        self._acctmp_doc.delete_one({'_creation_code': code})

        del acctmp['_id']
        del acctmp['_expired_date']
        del acctmp['_creation_code']

        result = self._acc_doc.insert_one(acctmp)
        del acctmp['hashpwd']

        return acctmp

    def find_id(self, id):
        return self._acc_doc.find_one({'_id': id})

    def find_name(self, name):
        return self._acc_doc.find_one({'name': name})

    def find_email(self, email):
        return self._acc_doc.find_one({'email': email})

    def remove(self, id):
        result = self._acc_doc.delete_one({'_id': id})
        if result.deleted_count != 1:
            raise AccountNotExist(id)

    def change_pwd(self, id, new_pwd):
        upd = {
            '$set': {
                'hashpwd': hash_pwd(new_pwd),
                'modified_date': datetime.utcnow(),
                'last_action': _ACT_CHANGE_PWD
            }
        }
        result = self._acc_doc.update_one({'_id': id}, upd)

        if result.modified_count != 1:
            raise AccountNotExist(id)

    def reset_pwd_code(self, email):
        acc = self._acc_doc.find_one({'email': email})
        if acc is None:
            raise AccountNotExist(email)
        self._rpwd_doc.delete_many({'acc_id': acc['_id']})

        reset_code = rand_code()
        code_spec = {
            'code': reset_code,
            'expired_date': time() + self.rpwd_time,
            'acc_id': acc['_id'],
            'acc_email': acc['email']
        }
        self._rpwd_doc.insert_one(code_spec)

        return reset_code

    def reset_pwd(self, code, new_pwd):
        code_spec = self._rpwd_doc.find_one({'code': code})
        if code_spec is None:
            raise CodeNotExistError(code)
        if code_spec['expired_date'] < time():
            raise CodeExpiredError()

        acc_id = code_spec['acc_id']
        self._rpwd_doc.delete_many({'acc_id': acc_id})

        new_hashpwd = hash_pwd(new_pwd)
        upd = {
            '$set': {
                'hashpwd': new_hashpwd,
                'last_action': _ACT_RESET_PWD
            }
        }
        self._acc_doc.update_one({'_id': acc_id}, upd)

        return acc_id

    def add_group(self, group_name):
        self._grp_doc.insert_one({'name': group_name})

    def del_group(self, group_name):
        result = self._grp_doc.delete_one({'name': group_name})

        if result.deleted_count != 1:
            raise GroupNotExist(group_name)

    def add_to_group(self, acc_id, group_name):
        acc = self._acc_doc.find_one({'_id': acc_id})
        if acc is None:
            raise AccountNotExist(acc_id)

        grp = self._grp_doc.find_one({'name': group_name})
        if grp is None:
            raise GroupNotExist(group_name)
        if group_name in acc['groups']:
            raise GroupExist(group_name)

        upd = {
            '$push': {'groups': group_name},
            '$set': {'last_action': _ACT_ADD_TO_GRP}
        }
        self._acc_doc.update_one({'_id': acc_id}, upd)

    def del_fm_group(self, acc_id, group_name):
        acc = self._acc_doc.find_one({'_id': acc_id})

        if acc is None:
            raise AccountNotExist(acc_id)
        if group_name not in acc['groups']:
            raise GroupNotExist(group_name)

        upd = {
            '$pull': {'groups': group_name},
            '$set': {'last_action': _ACT_RM_FRM_GRP}
        }
        self._acc_doc.update_one({'_id': acc_id}, upd)



from clink.com.type import Component


class AppErrHandler(Component):
    pass


class AppReqHandler(Component):
    pass


class AppResHandler(Component):
    pass


class Controller(Component):
    pass


class Service(Component):
    pass



'''
SYNOPSIS

    class MongoNode

DESCRIPTION

    instance() method must return connection.
    If connection wasn't established, create new connection.
    If connection was lost, reconnect and save new connection on success.

    doc() method must return document which specify by name.

    docs() method must return tuple of specify documents by names. if name
    of documents isn't in document specifies, raise error.

    close() method must close connection, release all of usage resources.

REFERENCES

    MongoDB - Wikipedia
        https://en.wikipedia.org/wiki/MongoDB
    PyMongo - APIs documents
        https://api.mongodb.com/python/current/index.html
'''

from pymongo import MongoClient, IndexModel

from clink.com import com, Component

from .error import DocumentNotExist, DocumentIndexError, DocSpecExit
from .type import MongoConf


@com(MongoConf)
class MongoService(Component):
    def __init__(self, conf):
        self._conf = conf
        self._doc_specs = []
        self._client = None
        self._doc_names = []

    def use_docspecs(self, doc_specs):
        for spec in doc_specs:
            self.use_docspec(spec)

    def use_docspec(self, doc_spec):
        db = self.instance()
        db_docs = db.collection_names()

        if doc_spec.name in db_docs:
            self._verify_spec(doc_spec)
        else:
            self._create_spec(doc_spec)

        self._doc_names.append(doc_spec.name)

    def _create_spec(self, doc_spec):
        db = self.instance()
        doc = db[doc_spec.name]

        if len(doc_spec.indexes) > 0:
            doc.create_indexes(doc_spec.indexes)

    def _verify_spec(self, doc_spec):
        # doc_sec: clink.db.MongoDocSpec
        # doc_index_info: result of pymongo.Collection.index_information()

        db = self.instance()
        doc_index_info = db[doc_spec.name].index_information()
        for index in doc_spec.indexes:
            # check index is exist
            if index.document['name'] not in doc_index_info:
                raise DocumentIndexError(doc_spec.name, index)

            # reconstruct index model from information
            index_name = index.document['name']
            doc_info = doc_index_info[index_name]
            doc_info_key = doc_info['key']
            doc_info['name'] = index_name
            del doc_info['key']
            doc_index = IndexModel(doc_info_key, **doc_info)

            # compare two index
            if not self._index_is_equal(index, doc_index):
                raise DocumentIndexError(doc_spec.name, index)

    def _index_is_equal(self, index_1, index_2):
        # index_1, index_2: pymongo.IndexModel
        doc_1 = index_1.document
        doc_2 = index_2.document
        for k in doc_1:
            if k not in doc_2:
                return False
            if doc_1[k] != doc_2[k]:
                return False
        return True

    def instance(self):
        self._connect()
        return self._client[self._conf.dbname]

    def doc(self, name):
        if name not in self._doc_names:
            raise DocumentNotExist(name)
        return self.instance()[name]

    def docs(self, *args):
        for name in args:
            if name not in self._doc_names:
                raise DocumentNotExist(name)
        db = self.instance()
        return tuple([db[name] for name in args])

    def close(self):
        if self._client is not None:
            self._client.close()

    def clear(self):
        self._connect()
        self._client.drop_database(self._conf.dbname)

    def _connect(self):
        if self._client is None:
            self._client = MongoClient(host=self._conf.dburl)



import json

from clink.error.http import Http400Error
from clink.iface import IPipeHandler
from clink.etc import UTF_8
from clink.mime.type import MIME_JSON
from clink.com.marker import com
from clink.type.com import AppReqHandler


@com()
class ReqJsonHandler(AppReqHandler, IPipeHandler):
    def handle(self, req, res):
        if req.content_type != MIME_JSON:
            return
        if req.body is None:
            return
        try:
            req.body = json.loads(req.body.decode(UTF_8))
        except ValueError:
            raise Http400Error(req, 'body is invalid json format')



class AccountExistError(Exception):
    def __init__(self, identity):
        self._msg = identity

    def __str__(self):
        return self._msg


class AccountNotExist(Exception):
    def __init__(self, identity):
        self._msg = identity

    def __str__(self):
        return self._msg


class GroupExist(Exception):
    def __init__(self, identity):
        self._msg = identity

    def __str__(self):
        return self._msg


class GroupNotExist(Exception):
    def __init__(self, identity):
        self._msg = identity

    def __str__(self):
        return self._msg


class PasswordError(Exception):
    def __init__(self, name, password):
        self._msg = 'usr=%s, pwd=%s' % (name, password)

    def __str__(self):
        return self._msg


class RTokenExpiredError(Exception):
    def __init__(self, rtoken_spec):
        self._msg = 'rtoken={}, created_date=[], expired_date={}'.format(
            rtoken_spec['value'], rtoken_spec['test_create'],
            rtoken_spec['expired_date']
        )

    def __str__(self):
        return self._msg


class TokenExpiredError(Exception):
    def __init__(self, atoken_raw):
        self._msg = 'sub=%s, exp=%s' % (atoken_raw['sub'], atoken_raw['exp'])

    def __str__(self):
        return self._msg


class CodeNotExistError(Exception):
    def __init__(self, code):
        self._msg = code

    def __str__(self):
        return self._msg


class CodeExpiredError(Exception):
    def __init__(self, spec):
        self._msg = 'code=%s, expired_date=%i'.format(
            spec['code'], spec['expired-date']
        )

    def __str__(self):
        return self._msg


class EmailExistError(Exception):
    def __init__(self, email):
        self._msg = email

    def __str__(self):
        return self._msg



'''
SYNOPSIS

    class OAuth

DESCRIPTION

    Parts of OAuth2 authentication and authorization. It doesn't support all
    of OAuth2 specification, here are supported:

        RFC 6749, section 4.3 -  Resource Owner Password Credentials Grant

        RFC 6749, section 6 - Refreshing an Access Token

        RFC 7519 - JSON Web Token

    Other specifications isn't supported because it's complicated without
    browsers. For example, mobile device need polls auth server to gets token
    instead of gets it from auth provider directly.

    Use this limited OAuth specification, you can perform external login
    with other OAuth Provider, you can only use name-password to get
    token and refresh that token. However, it work on all of platform.

    It also ignore authorization 'scope'. Authorization is perform by
    query database, not by information in access_token.

REFERENCES

    RFC 6749 - OAuth 2.0 Framework
        https://tools.ietf.org/html/rfc6749

    RFC 6750 - Bearer Token Usage
        https://tools.ietf.org/html/rfc6750

    RFC 6819 - Threat Model and Security Considerations
        https://tools.ietf.org/html/rfc6819

    RFC 7519 - JSON Web Token
        https://tools.ietf.org/html/rfc7519

    OAuth Home Page
        https://oauth.net/2/
'''

import jwt
from time import time
from bson import ObjectId

from clink.com import com
from clink.type.com import Service
from clink.service.mongo import MongoService
from clink.error.http import Http400Error, Http401Error

from .error import AccountNotExist, PasswordError, TokenExpiredError, \
                   RTokenExpiredError
from .authdb_sv import AuthDbService
from .type import AuthConf
from .util import hash_pwd


@com(MongoService, AuthDbService, AuthConf)
class OAuthService(Service):
    _TOKEN_ALG = 'HS512'

    def __init__(self, mongo_sv, authdb_sv, auth_conf):
        pass
        self._acc_doc = mongo_sv.doc('account')
        self._grp_doc = mongo_sv.doc('group')

        self._jwt_key = auth_conf.jwt_key
        self._token_time = auth_conf.token_time
        self._rtoken_time = auth_conf.rtoken_time

    def mktoken_pwd(self, name, password):
        acc = self._acc_doc.find_one({'name': name})
        if acc is None:
            raise AccountNotExist(name)

        hashpwd = hash_pwd(password)
        if hashpwd != acc['hashpwd']:
            raise PasswordError(name, password)

        return self._mktoken(acc['_id'])

    def mktoken_rtoken(self, rtoken):
        rtoken_raw = jwt.decode(
            rtoken, self._jwt_key, algorithm=self._TOKEN_ALG
        )
        if rtoken_raw['exp'] < time():
            raise RTokenExpiredError(rtoken_raw)

        return self._mktoken(rtoken_raw['sub'])

    def authen(self, access_token):
        atoken_raw = jwt.decode(
            access_token, self._jwt_key, algorithm=self._TOKEN_ALG
        )
        exp = int(atoken_raw['exp'])
        if exp < time():
            raise TokenExpiredError(atoken_raw)

        return ObjectId(atoken_raw['sub'])

    def _mktoken(self, acc_id):
        return {
            'token_type': 'Bearer',
            'expires_in': time() + self._token_time,
            'access_token': self._mkatoken(acc_id),
            'refresh_token': self._mkrtoken(acc_id)
        }

    def _mkatoken(self, acc_id):
        token_raw = {
            'sub': str(acc_id),
            'exp': time() + self._token_time
        }
        token = jwt.encode(
            token_raw, self._jwt_key, algorithm=self._TOKEN_ALG
        )
        return token.decode()

    def _mkrtoken(self, acc_id):
        token_raw = {
            'sub': str(acc_id),
            'exp': time() + self._rtoken_time
        }
        token = jwt.encode(
            token_raw, self._jwt_key, algorithm=self._TOKEN_ALG
        )
        return token.decode()

    def authen_req(self, req):
        if 'AUTHORIZATION' not in req.header:
            raise Http401Error(req)
        auth_header = req.header['AUTHORIZATION']
        auth_type = auth_header[:7]
        if auth_type != 'Bearer ':
            raise Http400Error(req)
        atoken = auth_header[7:]

        return self.authen(atoken)



'''
SYNOPSIS

    POST /auth/token

DESCRIPTION

    Create token with username, password or refresh_token
'''

from clink.com.marker import com
from clink.marker import route
from clink.type.com import Controller
from clink.service.auth import OAuthService


@com(OAuthService)
@route.path('auth')
class AuthCtl(Controller):
    def __init__(self, oauth_sv):
        self._oauth_sv = oauth_sv

    @route.post('token')
    def get_token(self, req, res):
        info = req.body
        grant_type = info['grant_type']

        if grant_type == 'password':
            res.body = self._oauth_sv.mktoken_pwd(
                info['username'], info['password']
            )
        elif grant_type == 'refresh_token':
            res.body = self._oauth_sv.mktoken_rtoken(info['refresh_token'])
        else:
            raise Http400Error(req, 'Not support grant_type=' + grant_type)



from clink.com.type import Component
from clink.com.marker import com


@com()
class AppConf(Component):
    def __init__(self, name, org_name=None, org_loc=None):
        self.name = name
        org_name = org_name
        org_loc = org_loc



'''
SYNOPSIS

    class ResJsonHandler

DESCRIPTION

    Encode object to JSON string. Use bson.json_util as encoder, output
    is strange. For example:

        o = {_id: ObjectId('5915e505e77989755d3cf4db')}

    After encode, result is:

        s = {"_id": {"$oid": "5915e505e77989755d3cf4db"}}
'''

from bson import json_util

from clink.iface import IPipeHandler
from clink.mime.type import MIME_JSON
from clink.etc import UTF_8
from clink.com.marker import com
from clink.type.com import AppResHandler


@com()
class ResJsonHandler(AppResHandler, IPipeHandler):
    def handle(self, req, res):
        if res.content_type != MIME_JSON:
            return
        if res.body is None:
            return
        res.body = json_util.dumps(res.body).encode(UTF_8)



import logging
from os import path
from os.path import isfile

from clink.etc import LOGFILE_MODE
from clink.iface import IPipeHandler
from clink.shell import touch
from clink.com.marker import com
from clink.com.type import Component
from clink.type import AppConf


@com(AppConf)
class ReqLogHandler(Component, IPipeHandler):
    def __init__(self, app_conf):
        file = path.join('/var/tmp', app_conf.name, 'request.log')
        if not isfile(file):
            touch(file, LOGFILE_MODE)

        self.logger = logging.getLogger(file)
        fhandler = logging.FileHandler(file)
        formatter = logging.Formatter('%(asctime)s$ %(message)s')
        fhandler.setFormatter(formatter)
        self.logger.addHandler(fhandler)
        self.logger.setLevel(logging.INFO)

    def handle(self, req, res):
        msg = '%s %s %s' % (req.remote_addr, req.method, req.path)
        self.logger.info(msg)



'''
SYSNOPSIS

    class Injector
        com_inst

        add_isnt(com_inst) 
        add_com(com_type)
        load()

DESCRIPTION

    com_inst is dictionary of instance of component. Key is type of component,
    value is isinstance of it. It must use as read only attribute, don't
    modify it.

    add_isnt() put an instance of component into list of instance of 
    component. com_inst argument must be subclass of clink.com.type.Component
    and marked as a component by clink.com.marker.com. add_isnt() only
    use to add configuration for other component because it doesn't support
    depedency component. It work like that for clear.

    add_com() put specification of component with com_type. com_type is
    type, not an instance. It must be subclass of clink.com.type.Component.
    Depedency components are specify via clink.com.marker.com.

    load() perform creating component by solve depedency components and
    put instance of depedency components to constructor.
'''

from collections import deque

from .type import Component
from .error import ComInvalidError, CircleComError, ComExistError, \
                   ComCreationError

CLINK_COM_ATTR = '__clink'


class Injector():
    def __init__(self):
        self._com_dict = {}
        self._com_layer = []
        self.com_inst = {}

    def add_inst(self, com_obj):
        com_type = type(com_obj)
        if not isinstance(com_obj, Component):
            raise ComInvalidError(com_type)
        if CLINK_COM_ATTR not in dir(com_type):
            raise ComInvalidError(com_type)
        if com_type in self._com_dict:
            raise ComExistError(com_type)
        clink_spec = getattr(com_type, CLINK_COM_ATTR)
        if len(clink_spec['req_coms']) > 0:
            raise InvalidDepedencyError(com_type)
        self._com_dict[com_type] = []
        self.com_inst[com_type] = com_obj

    def add_com(self, com_type):
        methods = dir(com_type)
        if CLINK_COM_ATTR not in methods:
            raise ComInvalidError(com_type)
        if com_type in self._com_dict:
            raise ComExistError(com_type)
        clink_spec = getattr(com_type, CLINK_COM_ATTR)
        self._com_dict[com_type] = clink_spec['req_coms']

    def add_coms(self, com_type):
        for t in com_type:
            self.add_com(t)

    def load(self):
        self._expand_com()
        self._mkcom_layer()
        self._mkcom_instance()

    def instance(self, type):
        if type not in self.com_inst:
            return None
        return self.com_inst[type]    

    def instanceof(self, com_type):
        coms = []
        for t in self.com_inst:
            if isinstance(self.com_inst[t], com_type):
                coms.append(self.com_inst[t])
        return coms

    def _expand_com(self):
        com_queue = deque(self._com_dict.keys())
        while len(com_queue) > 0:
            ct = com_queue.popleft()
            methods = dir(ct)
            if CLINK_COM_ATTR not in methods:
                raise ComInvalidError(ct)
            req_coms = getattr(ct, CLINK_COM_ATTR)['req_coms']
            if ct not in self._com_dict:
                self._com_dict[ct] = req_coms
            for t in req_coms:
                if t not in self._com_dict:
                    com_queue.append(t)

    def _mkcom_layer(self):
        com_list = list(self._com_dict.keys())
        while len(com_list) > 0:
            layer = []
            for ct in com_list:
                in_layer = True
                for dt in self._com_dict[ct]:
                    if dt in com_list:
                        in_layer = False
                        break
                if in_layer is True:
                    layer.append(ct)
                    com_list.remove(ct)
            if len(layer) == 0:
                raise CircleComError(com_list)
            self._com_layer.append(layer)

    def _mkcom_instance(self):
        for layer in self._com_layer:
            for com_type in layer:
                if com_type in self.com_inst:
                    continue
                arg_types = tuple([t for t in self._com_dict[com_type]])
                args = tuple([self.com_inst[t] for t in arg_types])
                try:
                    self.com_inst[com_type] = com_type(*args)
                except TypeError as e:
                    raise ComCreationError(com_type, args)



from smtplib import SMTP
from email.mime.text import MIMEText

from clink.com import com
from clink.type.com import Service
from clink.type import AppConf
from clink.service.auth import AuthConf


@com(AppConf, AuthConf)
class SmtpService(Service):
    def __init__(self, app_conf, auth_conf):
        self._app_conf = app_conf
        self._auth_conf = auth_conf

        self._from = '{} <{}>'.format(
            self._app_conf.name, self._auth_conf.root_email
        )

        self._smtp = self._connect()

    def send(self, dest_email, subject, text_body):
        msg = MIMEText(text_body)
        msg['To'] = dest_email
        msg['Subject'] = subject
        msg['From'] = self._from

        self._smtp.sendmail(
            self._auth_conf.root_email, dest_email, msg.as_string()
        )

    def _connect(self):
        smtp = SMTP(self._auth_conf.root_email_server)
        smtp.ehlo()
        smtp.starttls()
        smtp.login(self._auth_conf.root_email, self._auth_conf.root_email_pwd)

        return smtp



'''
SYNOPSIS

    POST /acc/reg/code
    POST /acc/reg

    GET  /acc/me
    PUT  /acc/me/pwd

    POST /acc/pwd/code
    POST /acc/pwd

DESCRIPTION

    Create code for registration. Confirm registration code to active account.

    Get information of account.

    Change password.

    Create code for reset password. Confirm reset password code and new
    password to reset password.
'''

from os import path
from os.path import dirname, realpath
from datetime import datetime, timedelta
from bson import ObjectId

from clink.error.http import Http401Error, Http400Error, Http500Error, \
                      Http404Error
from clink.service.auth import AccService, AuthConf
from clink.service.template import build_tpl
from clink.com.marker import com
from clink.marker import route
from clink.type.com import Controller
from clink.service.mail import SmtpService
from clink.service.auth import OAuthService
from clink.type import AppConf

_DIR = realpath(path.join(dirname(__file__), '../tpl'))
_REG_CODE_TMP_FILE = path.join(_DIR, 'reg-code.txt')
_REG_TMP_FILE = path.join(_DIR, 'reg-tmp.txt')
_CHANGE_PWD_TMP_FILE = path.join(_DIR, 'change-pwd.txt')
_RESET_PWD_TMP_FILE = path.join(_DIR, 'reset-pwd.txt')
_RESET_PWD_CODE_TMP_FILE = path.join(_DIR, 'reset-pwd-code.txt')


@com(AppConf, AuthConf, AccService, OAuthService, SmtpService)
@route.path('acc')
class AccountCtl(Controller):
    def __init__(self, app_conf, auth_conf, acc_sv, oauth_sv, smtp_sv):
        self._app_conf = app_conf
        self._auth_conf = auth_conf
        self._acc_sv = acc_sv
        self._oauth_sv = oauth_sv
        self._smtp_sv = smtp_sv

    @route.get('me')
    def get_me(self, req, res):
        acc_id = self._oauth_sv.authen_req(req)
        acc = self._acc_sv.find_id(acc_id)
        if acc is None:
            raise Http500Error(req, 'Account identity not found')

        res.body = {
            '_id': str(acc['_id']),
            'name': acc['name'],
            'email': acc['email'],
            'phone': acc['phone'],
            'created_date': int(acc['created_date'].timestamp()),
            'modifired_date': int(acc['modified_date'].timestamp()),
            'last_action': acc['last_action']
        }

    @route.post('reg/code')
    def create_reg_code(self, req, res):
        info = req.body

        reg_code = self._acc_sv.mk_creation(
            info['name'], info['pwd'], info['email']
        )

        datetime_now = datetime.utcnow()
        expired_date = datetime_now + timedelta(self._acc_sv.create_time)
        values = {
            'REG_CODE': reg_code,
            'APP_NAME': self._app_conf.name,
            'SENDER_NAME': 'root',
            'SENDER_EMAIL': self._auth_conf.root_email,
            'REMOTE_ADDR': req.remote_addr,
            'ACC_NAME': info['name'],
            'EMAIL': info['email'],
            'DATETIME_NOW': datetime_now.strftime('%Y-%m-%d %H:%M:%S'),
            'EXPIRED_DATE': expired_date.strftime('%Y-%m-%d %H:%M:%S')
        }
        txt_body = build_tpl(_REG_CODE_TMP_FILE, values)
        subject = 'Registration'
        self._smtp_sv.send(info['email'], subject, txt_body)

        res.status = 204

    @route.post('reg')
    def confirm_reg_code(self, req, res):
        reg_code = req.body['code']

        acc = self._acc_sv.confirm_creation(reg_code)

        datetime_now = datetime.utcnow()
        values = {
            'APP_NAME': self._app_conf.name,
            'SENDER_NAME': 'root',
            'SENDER_EMAIL': self._auth_conf.root_email,
            'REMOTE_ADDR': req.remote_addr,
            'ACC_NAME': acc['name'],
            'ACC_EMAIL': acc['email'],
            'DATETIME_NOW': datetime_now.strftime('%Y-%m-%d %H:%M:%S')
        }
        txt_body = build_tpl(_REG_TMP_FILE, values)
        subject = 'Registration'
        self._smtp_sv.send(acc['email'], subject, txt_body)

        res.status = 204

    @route.put('me/pwd')
    def change_pwd(self, req, res):
        new_pwd = req.body['new_pwd']

        acc_id = self._oauth_sv.authen_req(req)
        acc = self._acc_sv.find_id(acc_id)
        if acc is None:
            raise Http500Error(req, 'Account identity not found')
        self._acc_sv.change_pwd(acc_id, new_pwd)

        datetime_now = datetime.utcnow()
        values = {
            'APP_NAME': self._app_conf.name,
            'SENDER_NAME': 'root',
            'SENDER_EMAIL': self._auth_conf.root_email,
            'REMOTE_ADDR': req.remote_addr,
            'ACC_NAME': acc['name'],
            'DATETIME_NOW': datetime_now.strftime('%Y-%m-%d %H:%M:%S')
        }
        txt_body = build_tpl(_CHANGE_PWD_TMP_FILE, values)
        subject = 'Change password'
        self._smtp_sv.send(acc['email'], subject, txt_body)

        res.status = 204

    @route.post('pwd/code')
    def create_reset_pwd_code(self, req, res):
        email = req.body['email']

        acc = self._acc_sv.find_email(email)
        if acc is None:
            raise Http404Error(req, 'Email does not exist')

        reset_code = self._acc_sv.reset_pwd_code(email)
        expired_date = datetime.utcnow() + timedelta(hours=1)

        values = {
            'RESET_PWD_CODE': reset_code,
            'APP_NAME': self._app_conf.name,
            'SENDER_NAME': 'root',
            'SENDER_EMAIL': self._auth_conf.root_email,
            'REMOTE_ADDR': req.remote_addr,
            'DATETIME_NOW': datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S'),
            'EXPIRED_DATE': expired_date.strftime('%Y-%m-%d %H:%M:%S'),
            'ACC_NAME': acc['name']
        }
        txt_body = build_tpl(_RESET_PWD_CODE_TMP_FILE, values)

        subject = 'Reset password code'
        self._smtp_sv.send(email, subject, txt_body)

        res.status = 204

    @route.post('pwd')
    def confirm_reset_pwd_code(self, req, res):
        reset_code = req.body['code']
        new_pwd = req.body['new_pwd']

        acc_id = self._acc_sv.reset_pwd(reset_code, new_pwd)

        acc = self._acc_sv.find_id(acc_id)
        if acc is None:
            raise Http500Error(req)

        values = {
            'NEW_PWD': new_pwd,
            'RESET_PWD_CODE': reset_code,
            'ACC_NAME': acc['name'],
            'APP_NAME': self._app_conf.name,
            'REMOTE_ADDR': req.remote_addr,
            'DATETIME_NOW': datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S'),
            'SENDER_NAME': 'root',
            'SENDER_EMAIL': self._auth_conf.root_email,
        }
        txt_msg = build_tpl(_RESET_PWD_TMP_FILE, values)

        subject = 'Reset password'
        self._smtp_sv.send(acc['email'], subject, txt_msg)

        res.status = 204



from clink.com.injector import CLINK_COM_ATTR


def com(*args):
    def decorator_fn(target_obj):
        if CLINK_COM_ATTR in dir(target_obj):
            target_obj.__clink['req_coms'] = args
        else:
            target_obj.__clink = {'req_coms': args }
        return target_obj
    return decorator_fn



'''
SYNOPSIS

    class App
        add_com(com_type)
        load()
        __call__()

DESCRIPTION

    add_com() add component to application. com_type argument is type
    of component.

    load() create all of components, solve their depedency.

    __call__() is implements of WSGI.
'''

from os import path

from .iface import IWsgi
from .type import Request, Response
from .routing import Router
from .handler import RecvHandler, SendHandler
from .handler import ReqJsonHandler, ReqUrlEncodeHandler, ReqLogHandler
from .handler import ResJsonHandler, ResCorsHandler
from .handler import ErrorHttpHandler, ErrorLogHandler
from clink.com.injector import Injector, CLINK_COM_ATTR
from clink.type.com import Controller
from clink.routing import Route
from clink.type.com import AppErrHandler, AppReqHandler, AppResHandler


class App(IWsgi):
    req_log_handler = None
    err_log_handler = None

    err_handlers = None
    req_handlers = None
    res_handlers = None

    def __init__(self, conf):
        self.router = Router()
        self.injector = Injector()
        self.injector.add_inst(conf)

        self.add_com(RecvHandler)
        self.add_com(SendHandler)

        self.add_com(ReqLogHandler)
        self.add_com(ErrorLogHandler)
        self.add_com(ErrorHttpHandler)

        self.add_com(ReqJsonHandler)
        self.add_com(ReqUrlEncodeHandler)
        self.add_com(ResJsonHandler)
        self.add_com(ResCorsHandler)

    def add_com(self, com_type):
        self.injector.add_com(com_type)

    def load(self):
        self.injector.load()
        self._init_routes()

        self.recv_handler = self.injector.instance(RecvHandler)
        self.send_handler = self.injector.instance(SendHandler)
        self.req_log_handler = self.injector.instance(ReqLogHandler)
        self.err_log_handler = self.injector.instance(ErrorLogHandler)

        self.err_handlers = self.injector.instanceof(AppErrHandler)
        self.req_handlers = self.injector.instanceof(AppReqHandler)
        self.res_handlers = self.injector.instanceof(AppResHandler)

    def __call__(self, wsgi_env, wsgi_send):
        # level 0: recv handling
        req = Request()
        res = Response()

        try:
            # level 0 continue: recev handling
            self.recv_handler.handle(req, res, wsgi_env)

            # level 1: pre-routing handling
            self.req_log_handler.handle(req, res)

            # level 2: routing, find main handler
            m_req_handle = self.router.find_handle(req)

            # level 3: pre-main handling
            for handler in self.req_handlers:
                handler.handle(req, res)

            # level 4: main handling
            m_req_handle(req, res)

            # level 5: response handling
            for handler in self.res_handlers:
                handler.handle(req, res)

            # level 6: send handling
            return self.send_handler.handle(req, res, wsgi_send)
        except Exception as e:
            # level 7: error handling
            handled = False
            for handler in self.err_handlers:
                if handler.handle(req, res, e):
                    handled = True
            if not handled:
                raise e

            # level 8: error log handling
            self.err_log_handler.handle(req, res, e)

            # level 9: send error response
            return self.send_handler.handle(req, res, wsgi_send)

    def _init_routes(self):
        for type, obj in self.injector.com_inst.items():
            if isinstance(obj, Controller):
                self._add_ctl(obj)

    def _add_ctl(self, ctl):
        base_path = getattr(ctl, CLINK_COM_ATTR)['route_path']
        route = Route(base_path)

        for m in dir(ctl):
            mm = getattr(ctl, m)
            if CLINK_COM_ATTR in dir(mm):
                mm_attrs = getattr(mm, CLINK_COM_ATTR)
                if 'raw_route' not in mm_attrs:
                    continue
                raw_route = mm_attrs['raw_route']
                method = raw_route[0]
                path = raw_route[1]
                req_type = raw_route[2]
                route.add_spec(method, path, req_type, mm)

        self.router.add_route(route)



from clink.com import Component, com


@com()
class AuthConf(Component):
    def __init__(
        self, root_pwd, root_email, root_email_pwd, root_email_server,
        jwt_key, token_time=4*3600, rtoken_time=30*24*3600
    ):
        self.root_pwd = root_pwd
        self.root_email = root_email
        self.root_email_pwd = root_email_pwd
        self.root_email_server = root_email_server
        self.jwt_key = jwt_key
        self.token_time = token_time
        self.rtoken_time = rtoken_time



from clink.type.com import Service
from clink.com.marker import com
from clink.auth import Auth
from clink.ske.auth.type import AuthConfig


@com(AuthConfig)
class AuthService(Service):
    def __init__(self, auth_conf):
        pass
